/*                                                                             
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */

package org.wso2.solutions.identity.sts;

import java.io.ByteArrayInputStream;
import java.security.Key;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import javax.crypto.SecretKey;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.impl.dom.jaxp.DocumentBuilderFactoryImpl;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axis2.context.MessageContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.rahas.RahasConstants;
import org.apache.rahas.RahasData;
import org.apache.rahas.Token;
import org.apache.rahas.TokenIssuer;
import org.apache.rahas.TrustException;
import org.apache.rahas.TrustUtil;
import org.apache.ws.security.WSConstants;
import org.apache.ws.security.message.WSSecEncryptedKey;
import org.apache.ws.security.message.token.SecurityTokenReference;
import org.apache.ws.security.util.WSSecurityUtil;
import org.apache.ws.security.util.XmlSchemaDateFormat;
import org.apache.xml.security.encryption.EncryptedData;
import org.apache.xml.security.encryption.XMLCipher;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.signature.XMLSignature;
import org.apache.xml.security.utils.Base64;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLAttribute;
import org.opensaml.SAMLAttributeStatement;
import org.opensaml.SAMLException;
import org.opensaml.SAMLStatement;
import org.opensaml.SAMLSubject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.wso2.solutions.identity.IdentityConstants;
import org.wso2.solutions.identity.IdentityProviderException;
import org.wso2.solutions.identity.admin.ReportAdmin;
import org.wso2.solutions.identity.persistence.IPPersistenceManager;
import org.wso2.solutions.identity.persistence.dataobject.ActionDO;
import org.wso2.solutions.identity.persistence.dataobject.InfoCardDO;
import org.wso2.solutions.identity.persistence.dataobject.IssuedTokensDO;
import org.wso2.solutions.identity.persistence.dataobject.PPIDValueDO;
import org.wso2.solutions.identity.persistence.dataobject.RelyingPartyDO;
import org.wso2.solutions.identity.sts.IdentityProviderData.RequestedClaimData;
import org.wso2.utils.ServerConfiguration;

/**
 * This is the main token issuer implementation identity provider.
 * Two types of credential mechanisms are supported:
 * 
 * <ul><li>Username and Password Credential</li>
 * <li>Self-issued Token Credential</li>
 * </ul>
 * 
 * Supported list of claims can be configured and the values of those supported 
 * claims will be obtained from the user store setup to work with the 
 * identity provider.
 */
public class IdentityTokenIssuer implements TokenIssuer {

    private static Log log = LogFactory.getLog(IdentityTokenIssuer.class);

    private final static String WSS_SAML_NS = "http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.0#";

    String confileFilePath;

    IdentityProviderData ipData;

    public final static String ISSUER_SELF = IdentityConstants.NS
            + "/issuer/self";

    /**
     * {@inheritDoc}
     */
    public String getResponseAction(RahasData data) throws TrustException {
        return RahasConstants.WST_NS_05_02 + RahasConstants.RSTR_ACTION_ISSUE;
    }

    /**
     * {@inheritDoc}
     */
    public SOAPEnvelope issue(RahasData data) throws TrustException {

        boolean debug = log.isDebugEnabled();

        if (debug) {
            log.debug("issue");
        }
        try {
            if (debug) {
                log.debug("Request: \n" + data.getRstElement().toString()
                        + "\n\n");
            }

            ipData = new IdentityProviderData(data);
            boolean isValidCard = isValidCard(ipData.getCardID());

            if (debug) {
                log.debug("Card is validated");
            }

            if (isValidCard == false) {
                String msg = "Invalid information card";
                log.error(msg);
                ReportAdmin admin = new ReportAdmin();
                admin.record(ipData.getUserIdentifier(),
                        ActionDO.ACTION_TOKEN_ISSUE_FAILURE, msg);
                throw new TrustException(TrustException.REQUEST_FAILED);

            }

            return createResponse(data);

        } catch (Exception e) {
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        } finally {
            log.info("Issued token");
        }

    }

    /**
     * Create the response SOAP envelope.
     * @param data WS-Trust information in the issue request.
     * @return response SOAP envelope.
     * @throws TrustException
     */
    private SOAPEnvelope createResponse(RahasData rahasData)
            throws TrustException {

        try {

            Date notBefore = new Date();
            Date notAfter = new Date();
            notAfter.setTime(notBefore.getTime() + (300 * 1000));

            // Set the DOM impl to DOOM
            DocumentBuilderFactoryImpl.setDOOMRequired(true);

            MessageContext inMsgCtx = rahasData.getInMessageContext();

            SOAPEnvelope env = TrustUtil.createSOAPEnvelope(inMsgCtx
                    .getEnvelope().getNamespace().getNamespaceURI());
            Document doc = ((Element) env).getOwnerDocument();

            // Create EncryptedKey
            WSSecEncryptedKey encryptedKey = null;

            X509Certificate serviceCert = ipData.getRpCert();
            if (serviceCert != null) {
                encryptedKey = new WSSecEncryptedKey();
                encryptedKey.setUseThisCert(serviceCert);
                encryptedKey.setKeySize(256);
                encryptedKey.setKeyEncAlgo(WSConstants.KEYTRANSPORT_RSAOEP);
                encryptedKey
                        .setKeyIdentifierType(WSConstants.THUMBPRINT_IDENTIFIER);
                encryptedKey.prepare(doc, null);

                Element encrKeyElem = encryptedKey.getEncryptedKeyElement();

                // Create KeyInfo structure - START

                Element keyInfoElem = doc.createElementNS(WSConstants.SIG_NS,
                        "KeyInfo");
                ((OMElement) encrKeyElem).declareNamespace(WSConstants.SIG_NS,
                        WSConstants.SIG_PREFIX);
                ((OMElement) encrKeyElem).declareNamespace(WSConstants.ENC_NS,
                        WSConstants.ENC_PREFIX);
                keyInfoElem.appendChild(encrKeyElem);

                // Create KeyInfo structure - END

            }

            SAMLStatement[] statements = createSAMLStatements(ipData, rahasData);

            // Assertion
            ServerConfiguration config = ServerConfiguration.getInstance();
            String host = "http://" + config.getFirstProperty("HostName");
            SAMLAssertion assertion = new SAMLAssertion(host, notBefore,
                    notAfter, null, null, Arrays.asList(statements));

            // Sign the assertion
            String keyAlias = ServerConfiguration.getInstance()
                    .getFirstProperty("Security.KeyStore.KeyAlias");
            X509Certificate[] issuerCerts = KeyUtil
                    .getServiceCertificateChain(keyAlias);

            String sigAlgo = XMLSignature.ALGO_ID_SIGNATURE_RSA;
            String pubKeyAlgo = issuerCerts[0].getPublicKey().getAlgorithm();
            if (pubKeyAlgo.equalsIgnoreCase("DSA")) {
                sigAlgo = XMLSignature.ALGO_ID_SIGNATURE_DSA;
            }
            Key issuerPK = KeyUtil.getPrivateKey(keyAlias);
            assertion.sign(sigAlgo, issuerPK, Arrays.asList(issuerCerts));

            // This will be required to send back to the requester in the RSTR
            OMElement rstrElem = createRSTR(rahasData, notBefore, notAfter,
                    env, doc, assertion, encryptedKey);

            if (log.isDebugEnabled()) {
                log.debug("Response created");
                log.debug("Response body : \n" + rstrElem.toString() + "\n\n");
            }
            IPPersistenceManager dbman = IPPersistenceManager
                    .getPersistanceManager();
            InfoCardDO card = dbman.getInfoCard(ipData.getCardID());

            IssuedTokensDO tok = new IssuedTokensDO();
            tok.setCard(card);
            tok.setDateExpires(notAfter);
            tok.setDateIssued(notBefore);
            tok.setTokenType(rahasData.getTokenType());

            dbman.create(tok);
            return env;

        } catch (Exception e) {
            log.error(e.getMessage());
            try {
                ReportAdmin.record(ipData.getUserIdentifier(), ActionDO.ACTION_TOKEN_ISSUE_FAILURE, e.getMessage());
            } catch (IdentityProviderException e1) {
                throw new TrustException(TrustException.REQUEST_FAILED, e1);
            }
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        } finally {
            // Un-set the DOM impl to default
            DocumentBuilderFactoryImpl.setDOOMRequired(false);
            log.info("Response ready for : " + ipData.getCardID());
        }
    }

    /**
     * Create the <code>wst:RequstedSecurityTokenRespoonse</code> element.
     * @param data WS-Trust information in the issue request
     * @param notBefore Created time
     * @param notAfter Expiration time
     * @param env Response SOAP envelope
     * @param doc <code>org.w3.dom.Document</code> instance of the response SOAP envelope
     * @param assertion SAML Assertion to be sent in the response.
     * @param encryptedKey Key used to encrypt the SAML assertion.
     * @return <code>wst:RequstedSecurityTokenRespoonse</code> element.
     * @throws TrustException
     * @throws SAMLException
     */
    private OMElement createRSTR(RahasData data, Date notBefore, Date notAfter,
            SOAPEnvelope env, Document doc, SAMLAssertion assertion,
            WSSecEncryptedKey encryptedKey) throws TrustException,
            SAMLException, IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin RSTR Element creation.");
        }
        int wstVersion = data.getVersion();
        MessageContext inMsgCtx = data.getInMessageContext();

        OMElement rstrElem = TrustUtil
                .createRequestSecurityTokenResponseElement(wstVersion, env
                        .getBody());
        TrustUtil.createTokenTypeElement(wstVersion, rstrElem).setText(
                data.getTokenType());

        createDisplayToken(rstrElem, ipData);

        OMElement appliesToEpr = null;
        if (encryptedKey != null) {
            int keysize = data.getKeysize();
            if (keysize == -1) {
                keysize = encryptedKey.getEphemeralKey().length * 8;
            }

            TrustUtil.createKeySizeElement(wstVersion, rstrElem, keysize);

            OMElement incomingAppliesToEpr = data.getAppliesToEpr();
            try {
                Document eprDoc = DocumentBuilderFactory.newInstance()
                        .newDocumentBuilder().parse(
                                new ByteArrayInputStream(incomingAppliesToEpr
                                        .toString().getBytes()));
                appliesToEpr = (OMElement) doc.importNode(eprDoc
                        .getDocumentElement(), true);
            } catch (Exception e) {
                new TrustException(TrustException.REQUEST_FAILED, e);
            }

            OMElement appliesToElem = rstrElem
                    .getOMFactory()
                    .createOMElement(
                            new QName(
                                    RahasConstants.WSP_NS,
                                    RahasConstants.IssuanceBindingLocalNames.APPLIES_TO,
                                    RahasConstants.WSP_PREFIX), rstrElem);
            appliesToElem.addChild(appliesToEpr);
        }

        // Use GMT time in milliseconds
        DateFormat zulu = new XmlSchemaDateFormat();

        // Add the Lifetime element
        TrustUtil.createLifetimeElement(wstVersion, rstrElem, zulu
                .format(notBefore), zulu.format(notAfter));

        OMElement reqSecTokenElem = TrustUtil
                .createRequestedSecurityTokenElement(wstVersion, rstrElem);

        Node tempNode = assertion.toDOM();
        Node assertionElement = doc.importNode(tempNode, true);
        reqSecTokenElem.addChild((OMNode) assertionElement);

        if (log.isDebugEnabled()) {
            log.debug(assertionElement.toString());
        }

        if (encryptedKey != null) {
            encryptSAMLAssertion(doc, (Element) assertionElement, encryptedKey);
        }

        createAttachedRef(rstrElem, assertion.getId());
        createUnattachedRef(rstrElem, assertion.getId());

        // Store the Token
        Token assertionToken = new Token(assertion.getId(), (OMElement) doc
                .importNode(assertion.toDOM(), true), notBefore, notAfter);

        // At this point we definitely have the secret
        // Otherwise it should fail with an exception earlier
        assertionToken.setSecret(data.getEphmeralKey());
        TrustUtil.getTokenStore(inMsgCtx).add(assertionToken);

        // Creating the ReqProoftoken - END
        if (log.isDebugEnabled()) {
            log.debug("RSTR Elem created.");
        }
        log.info("RSTR ready with token : " + assertion.getId());
        return rstrElem;
    }

    /**
     * Create and add wst:AttachedReference element
     * @param rstrElem wst:RequestSecurityToken element
     * @param id Token identifier
     */
    private void createAttachedRef(OMElement rstrElem, String id) {
        OMFactory fac = rstrElem.getOMFactory();
        OMElement rar = fac
                .createOMElement(
                        new QName(
                                RahasConstants.WST_NS_05_02,
                                RahasConstants.IssuanceBindingLocalNames.REQUESTED_ATTACHED_REFERENCE,
                                RahasConstants.WST_PREFIX), rstrElem);
        OMElement str = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                SecurityTokenReference.SECURITY_TOKEN_REFERENCE,
                WSConstants.WSSE_PREFIX), rar);
        OMElement ki = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                "KeyIdentifier", WSConstants.WSSE_PREFIX), str);
        ki.addAttribute("ValueType", WSS_SAML_NS
                + WSConstants.SAML_ASSERTION_ID, null);
        ki.setText(id);
    }

    /**
     * Create and add wst:UnattachedReference element
     * @param rstrElem wst:RequestSecurityToken element
     * @param id Token identifier
     */
    private void createUnattachedRef(OMElement rstrElem, String id) {
        OMFactory fac = rstrElem.getOMFactory();
        OMElement rar = fac
                .createOMElement(
                        new QName(
                                RahasConstants.WST_NS_05_02,
                                RahasConstants.IssuanceBindingLocalNames.REQUESTED_UNATTACHED_REFERENCE,
                                RahasConstants.WST_PREFIX), rstrElem);
        OMElement str = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                SecurityTokenReference.SECURITY_TOKEN_REFERENCE,
                WSConstants.WSSE_PREFIX), rar);
        OMElement ki = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                "KeyIdentifier", WSConstants.WSSE_PREFIX), str);

        ki.addAttribute("ValueType", WSS_SAML_NS
                + WSConstants.SAML_ASSERTION_ID, null);
        ki.setText(id);
    }

    /**
     * Encrypt the given SAML Assertion element with the given key information.
     * @param doc
     * @param assertionElement
     * @param encryptedKey
     */
    private void encryptSAMLAssertion(Document doc, Element assertionElement,
            WSSecEncryptedKey encryptedKey) throws TrustException {
        try {
            XMLCipher xmlCipher = XMLCipher.getInstance(WSConstants.AES_256);
            SecretKey secretKey = WSSecurityUtil.prepareSecretKey(
                    WSConstants.AES_256, encryptedKey.getEphemeralKey());
            xmlCipher.init(XMLCipher.ENCRYPT_MODE, secretKey);
            String xencEncryptedDataId = "EncDataId-"
                    + assertionElement.hashCode();

            KeyInfo keyInfo = new KeyInfo(doc);
            keyInfo.addUnknownElement(encryptedKey.getEncryptedKeyElement());

            EncryptedData encData = xmlCipher.getEncryptedData();
            encData.setId(xencEncryptedDataId);
            encData.setKeyInfo(keyInfo);
            xmlCipher.doFinal(doc, assertionElement, false);
        } catch (Exception e) {
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        }
    }

    /**
     * Create the DisplayToken element according to CardSpace specifications.
     * @param rahasData Information from the WS-Trust request.
     * @param ipData CardSpace specific meta-data for this issuance.
     * @return The DisplayToken element. 
     */
    private OMElement createDisplayToken(OMElement rstrElem,
            IdentityProviderData ipData) throws IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin Display token creation.");
        }

        Map requestedClaims = ipData.getRequestedClaims();

        if (requestedClaims.isEmpty())
            return null;

        OMElement rdt = IdentityProviderUtil.createRequestedDisplayToken(
                rstrElem, ipData);
        OMElement displayToken = IdentityProviderUtil.createDisplayToken(rdt,
                ipData);

        Iterator ite = requestedClaims.values().iterator();
        while (ite.hasNext()) {
            RequestedClaimData claim = (RequestedClaimData) ite.next();

            if (claim.uri.equals(IdentityConstants.CLAIM_PPID)) {
                //PPID display token
                IdentityProviderUtil.createDisplayClaim(displayToken, ipData
                        .getDisplayName(claim.uri), IdentityProviderUtil
                        .getPPIDDisplayValue(claim.value), claim.uri);
            } else {
                IdentityProviderUtil.createDisplayClaim(displayToken, ipData
                        .getDisplayName(claim.uri), claim.value, claim.uri);
            }
        }

        if (log.isDebugEnabled()) {
            log.debug("createDisplayToken");
        }

        return rdt;
    }

    /**
     * Create SAML attribute statements for the response token.
     * @param ipData Meta-data for this issuer.
     * @param rahasData WS-Trust common information in the issue request.
     * @return An array of SAML attribute statements
     * @throws SAMLException
     * @throws IdentityProviderException
     */
    private SAMLStatement[] createSAMLStatements(IdentityProviderData ipData,
            RahasData rahasData) throws SAMLException,
            IdentityProviderException {

        // The SAMLSubject
        if (log.isDebugEnabled()) {
            log.debug("Begin SAML statement creation.");
        }
        String[] confirmationMethods = new String[] { SAMLSubject.CONF_HOLDER_KEY };

        SAMLSubject subject = new SAMLSubject(null, Arrays
                .asList(confirmationMethods), null, null);

        Map mapClaims = ipData.getRequestedClaims();

        List assertionList = new ArrayList();
        Iterator ite = mapClaims.values().iterator();

        while (ite.hasNext()) {
            RequestedClaimData claim = (RequestedClaimData) ite.next();
            String uri = claim.uri;
            if (uri.equals(IdentityConstants.CLAIM_PPID)) {
                // Get this user's PPID
                String name = ipData.getUserIdentifier();
                claim.value = getPPID(rahasData, name, rahasData
                        .getAppliesToEpr());
            }

            int index = uri.lastIndexOf("/");

            if(claim.value == null) {
                throw new IdentityProviderException(
                        "noValueForRequestedAttribute",
                        new String[] { claim.uri });
            }
            List list = Arrays.asList(new String[] { claim.value });
            String displayTag = ipData.getDisplayName(uri);
            if ((displayTag == null) && (claim.bOptional == false)) {
                throw new IdentityProviderException("unknownClaimUri",
                        new String[] { uri });
            }
            SAMLAttribute attr = new SAMLAttribute(uri.substring(index + 1, uri
                    .length()), uri.substring(0, index) , null, -1,
                    list);
            assertionList.add(attr);

        }
        SAMLAttribute[] attrs = (SAMLAttribute[]) assertionList
                .toArray(new SAMLAttribute[assertionList.size()]);

        // AttributeStatement
        SAMLAttributeStatement attrStmt = new SAMLAttributeStatement(subject,
                Arrays.asList(attrs));

        SAMLStatement[] statements = { attrStmt };
        if (log.isDebugEnabled()) {
            log.debug("SAML statements created");
        }
        return statements;
    }

    /**
     * Obtain the ppid for the given user for the given rp.
     * @param rahasData WS-Trust information in the issue request.
     * @param name Name of the user/subject.
     * @param appliesToEpr EPR element in wst:AppliesTo element.
     * @return PPID value. If there's already an issued token then the ppid 
     * value will be reused. 
     * @throws IdentityProviderException
     */
    private String getPPID(RahasData rahasData, String name,
            OMElement appliesToEpr) throws IdentityProviderException {

        String appliesToHostName = IdentityProviderUtil
                .getAppliesToHostName(rahasData);
        IPPersistenceManager db = IPPersistenceManager.getPersistanceManager();
        PPIDValueDO[] ppidValueDOs = db.getPPIDValuesForUser(name);
        PPIDValueDO ppidValueDO = null;
        for (int i = 0; i < ppidValueDOs.length; i++) {
            String hostName = null;
            if (ppidValueDOs[i].getRelyingParty() != null) {
                hostName = ppidValueDOs[i].getRelyingParty().getHostName();
            } else if (ppidValueDOs[i].getPersonalRelyingParty() != null) {
                hostName = ppidValueDOs[i].getPersonalRelyingParty()
                        .getIdentifier().getHostName();
            }

            if (hostName == null) {
                throw new IdentityProviderException("hostNotTrusted",
                        new String[] { hostName });
            }

            if (appliesToHostName.equals(hostName)) {
                ppidValueDO = ppidValueDOs[i];
            }
        }

        if (ppidValueDO != null) {
            //If we have already issued a PPID 
            //Then return that value
            return ppidValueDO.getPpid();
        } else {
            //A new request targeted for a new RP
            String newPpid = Base64.encode(UUIDGenerator.getUUID().getBytes());
            ppidValueDO = new PPIDValueDO();
            ppidValueDO.setUserId(name);
            ppidValueDO.setPpid(newPpid);

            //If the host is globally trusted
            RelyingPartyDO rp = db.getRelyingParty(appliesToHostName);
            if (rp != null) {
                ppidValueDO.setRelyingParty(rp);
            } else {
                //Else the host MUST be personally trusted
                ppidValueDO.setPersonalRelyingParty(db.getPersonalRelyingParty(
                        name, appliesToHostName));
            }

            db.create(ppidValueDO);

            return newPpid;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationElement(OMElement configElement) {
        //Nothing to do
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationFile(String configFile) {
        this.confileFilePath = configFile;
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationParamName(String configParamName) {
        //Nothing to do
    }

    /**
     * Check whether the information card referenced in the token request is 
     * a valid managed information card issued by the identity provider.
     * @param cardId Identifier of the information card.
     * @return Whether the card is valid or not.
     * @throws IdentityProviderException
     */
    private boolean isValidCard(String cardId) throws IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin Validating card.");
        }
        boolean retval = false;
        IPPersistenceManager dbman = IPPersistenceManager
                .getPersistanceManager();
        InfoCardDO card = dbman.getInfoCard(cardId);
        if (card != null) {
            Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
            Date now = cal.getTime();

            if (now.before(card.getDateExpires())
                    && now.after(card.getDateIssued())) {
                retval = true;
            }
        }

        return retval;
    }

}
